import sys
import signal
import logging
from os import utime
from time import time
from random import randint
from .api import send
from .data import Data
from . import timestamp
from .libvirt import Libvirt, event_listener

log = logging.getLogger(__name__)


class Daemon:
    def __init__(self) -> None:
        log.info('Daemon initializing')

        # Add all the signals that should be handled to this dict.
        # Use the following format:
        #   signal number: ('signal name', 'short description')
        self.sig_mapping = {
            # Systemd stop/restart sends SIGTERM by default.
            15: ('SIGTERM', 'exiting gracefully')
        }

        for item in self.sig_mapping.items():
            signal.signal(item[0], self.sig_handler)
            log.debug(f'Configured handler for {item[1][0]}')

        # Create a timestamp file if one does not exist, but prevent
        # the modification timestamp from being overwritten when the service is restarted.
        if not timestamp.is_file():
            timestamp.touch()
            # Set the modification timestamp to the beginning of the epoch
            # to ensure that the data is sent after a fresh installation or hypervisor reboot.
            utime(timestamp.as_posix(), (1, 1))

        # This random number is used when checking the age of
        # the timestamp to avoid many hosts sending data at the same time.
        self.rand_num = randint(79200, 86400)

    def sig_handler(self, sig_num, frame) -> None:
        # Handle individual signals in this function.
        log.info(f'Daemon recived {self.sig_mapping[sig_num][0]}, {self.sig_mapping[sig_num][1]}')

        # SIGTERM
        if sig_num == 15:
            sys.exit(0)

        else:
            log.critical(f'Handler has no instructions on how to handle signal number {sig_num}')
            sys.exit(1)

    def daemon(self) -> None:
        # Open the libvirt connection.
        with Libvirt() as conn:
            log.info('Daemon started')

            while True:
                # Listen for libvirt events.
                event_listener()

                # Compare modification timestamp with current time
                # and trigger send if timestamp is older than ~24 hours.
                if time() > timestamp.stat().st_mtime + self.rand_num:
                    log.info('Data was sent more than 24 hours ago, or the hypervisor was rebooted, triggering send')
                    send(api_data=Data(conn).main())
