import re
import sys
import yaml
import logging
from glob import glob
from json import dumps
from libvirt import virConnect, libvirtError
from ipaddress import IPv4Interface, IPv6Interface
from configparser import ConfigParser, NoOptionError, NoSectionError
from . import conf_dir
from .helpers import load_config

log = logging.getLogger(__name__)


# This class returns the host and VM data as a dictionary.
class Data(object):
    def __init__(self, conn: virConnect) -> None:
        log.debug('Gather data')

        self.conn = conn
        self.data = {
            'virtualization': 'kvm',
            'hardware': {},
            'vms': []
        }

        self.host()
        self.vm()

    def host(self) -> None:
        # Environment
        log.debug('Get environment')
        try:
            with open('/etc/ansible/facts.d/machine_environment.fact', 'r') as f:
                fact = f.read()
                if re.search(r'prod', fact):
                    log.debug('Production environment detected')
                    self.data['environment'] = 'prod'
                elif re.search(r'dev', fact):
                    log.debug('Development environment detected')
                    self.data['environment'] = 'dev'
                else:
                    log.critical(f'Unknown environment: {fact}')
                    sys.exit(1)
        except IOError as e:
            log.critical(f'Could not read environment file: {e}')
            sys.exit(1)

        # CPU
        # This will give you the amount of CPU cores, not threads!
        log.debug('Get host_cpu')
        with open('/proc/cpuinfo', 'r') as f:
            host_cpu = int(re.search(r'cpu\scores\s+:\s(\d+)', f.read()).groups()[0])
        log.debug(f'{host_cpu=}')

        # RAM
        # All 'meminfo' values are in KiB, multiply by 1024 to get bytes.
        log.debug('Get host_ram')
        with open('/proc/meminfo', 'r') as f:
            host_ram = int(re.search(r'^MemTotal:\s+(\d+)\skB', f.read()).groups()[0]) * 1024
        log.debug(f'{host_ram=}')

        # Disk
        # All md-size values are in 512-byte blocks, multiply by 512 to get bytes.
        log.debug('Get host_disk')
        host_disk = 0
        for f in glob('/sys/block/md*/size', recursive=True):
            with open(f, 'r') as value:
                host_disk += int(value.read()) * 512
        log.debug(f'{host_disk=}')

        # JSON
        host_data = {
            'cpu': host_cpu,
            'ram': host_ram,
            'disk': host_disk
        }
        self.data['hardware'].update(host_data)

        # inventory-client.d
        # If you need to manually override host data,
        # this can be done using a host.ini file placed in the {conf_dir} directory.
        log.debug('Looking for host override file')
        config = load_config(path=f'{conf_dir}/host.ini', fatal=False, parser=ConfigParser())
        if config:
            log.debug('Override file found, checking content')

            # Overrides defined in this dictonray will be searched for
            # and used to override data if found in the override file.
            overrides = {
                'environment': 'main',
                'virtualization': 'main'
            }

            for option, section in overrides.items():
                try:
                    self.data[option] = config.get(section, option)
                    log.debug(f"'{option}' successfully overridden")

                except (NoSectionError, NoOptionError) as e:
                    log.debug(f"Could not override '{option}': {e}")

        del config

    def vm(self) -> None:
        try:
            for dom in self.conn.listAllDomains():
                # Name
                # In the context of libvrit, the "name()" is the order number (Auftragsnummer).
                log.debug('Get vm_name')
                vm_name = str(dom.name())
                log.debug(f'{vm_name=}')

                # FQDN
                # AFAIK, the only way to get the FQDN of a VM (even if it is off) is to parse its XML configuration.
                log.debug('Get vm_fqdn')
                vm_fqdn = re.search(r'<title>.*?(\S+\.\S+\.\S+)</title>', dom.XMLDesc()).groups()[0]
                log.debug(f'{vm_fqdn=}')

                # IPv4/6
                log.debug('Get vm_ipv4/6')
                try:
                    with open(f'/etc/grommet/{vm_name}.yml', 'r') as f:
                        ip_data = yaml.safe_load(f)

                except IOError as e:
                    log.critical(f'Could not read network config: {e}')
                    sys.exit(1)

                vm_ipv4 = str(IPv4Interface(ip_data['routes']['vmbr0']['v4'][0]).ip)
                log.debug(f'{vm_ipv4=}')

                # VMs do not currently have IPv6, do not throw error if it's missing.
                try:
                    vm_ipv6 = str(IPv6Interface(ip_data['routes']['vmbr0']['v6'][0]).ip)
                except KeyError:
                    vm_ipv6 = 'null'
                log.debug(f'{vm_ipv6=}')

                # CPU
                # This would give you the amount of vCPUs (threads) available to a VM.
                log.debug('Get vm_cpu')
                vm_cpu = int(dom.info()[3])
                log.debug(f'{vm_cpu=}')

                # RAM
                # This value is in KiB, multiply by 1024 to get bytes.
                log.debug('Get vm_ram')
                vm_ram = int(dom.info()[2]) * 1024
                log.debug(f'{vm_ram=}')

                # Disk
                # Libvirt's storageVolLookup returns bytes.
                log.debug('Get vm_disk')
                pool = self.conn.storagePoolLookupByName('vgvm')
                vm_disk = int(pool.storageVolLookupByName(f'{vm_name}-root').info()[2])
                log.debug(f'{vm_disk=}')

                # JSON
                vm_data = {
                    'name': vm_name,
                    'fqdn': vm_fqdn,
                    'ipv4': vm_ipv4,
                    'ipv6': vm_ipv6,
                    'hardware': {
                        'cpu': vm_cpu,
                        'ram': vm_ram,
                        'disk': vm_disk
                    }
                }

                # inventory-client.d
                # If you need to manually override data for a specific VM,
                # this can be done using a {vm_name}.ini file placed in the {conf_dir} directory.
                log.debug(f'Looking for {vm_name} override file')
                config = load_config(path=f'{conf_dir}/{vm_name}.ini', fatal=False, parser=ConfigParser())
                if config:
                    log.debug('Override file found, checking content')

                    # Overrides defined in this dictonray will be searched for
                    # and used to override data if found in the override file.
                    overrides = {
                        'group': 'main'
                    }

                    for option, section in overrides.items():
                        try:
                            vm_data[option] = config.get(section, option)
                            log.debug(f"'{option}' successfully overridden")

                        except (NoSectionError, NoOptionError) as e:
                            log.debug(f"Could not override '{option}': {e}")

                del config

                self.data['vms'].append(vm_data)

        except libvirtError as e:
            log.critical(f'Could not gather data: {e}')
            sys.exit(1)

    def main(self) -> dict:
        log.debug(dumps(self.data, indent=4))
        return self.data
