import sys
import libvirt
import logging
from .api import send
from .data import Data

log = logging.getLogger(__name__)


# To listen for events, use this function in a loop.
def event_listener():
    try:
        libvirt.virEventRunDefaultImpl()

    except libvirt.libvirtError as e:
        log.critical(f'Event listener failed: {e}')
        sys.exit(1)


class Libvirt:
    def __init__(self) -> None:
        # Address to connect to.
        self.addr = 'qemu:///system'

        # Map domain event names to their corresponding numbers.
        self.domevent_mapping = {
            0: 'VIR_DOMAIN_EVENT_DEFINED',
            1: 'VIR_DOMAIN_EVENT_UNDEFINED',
            2: 'VIR_DOMAIN_EVENT_STARTED',
            3: 'VIR_DOMAIN_EVENT_SUSPENDED',
            4: 'VIR_DOMAIN_EVENT_RESUMED',
            5: 'VIR_DOMAIN_EVENT_STOPPED',
            6: 'VIR_DOMAIN_EVENT_SHUTDOWN',
            7: 'VIR_DOMAIN_EVENT_PMSUSPENDED',
            8: 'VIR_DOMAIN_EVENT_CRASHED',
            9: 'VIR_DOMAIN_EVENT_LAST'
        }

    def __enter__(self) -> libvirt.virConnect:
        log.debug(f'Opening libvirt connection ({self.addr})')

        try:
            # Register the default event implementation. (must be done before opening the connection)
            libvirt.virEventRegisterDefaultImpl()

            # This connection type has a restricted set of method calls that are
            # allowed and is useful for applications that should not be allowed to make changes.
            self.conn = libvirt.openReadOnly(self.addr)
            
            self.domevents = [
                # Configure the type of event to receive and specify the handling function.
                self.conn.domainEventRegisterAny(None, libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE, self.domevent_handler, None)
            ]

            return self.conn

        except libvirt.libvirtError as e:
            log.critical(f'Could not connect: {e}')
            sys.exit(1)

    def domevent_handler(self, conn, dom, event, detail, opaque) -> None:
        try:
            log.info(f'Received {self.domevent_mapping[event]} ({dom.name()})')
        except KeyError:
            log.error(f'Received unmapped event number {event} ({dom.name()})')

        # Handle individual domain events in here.
        if event == 0 or event == 1 or event == 2:
            log.info('Triggering send')
            send(api_data=Data(self.conn).main())

        else:
            log.info('Irrelevenat event, ignoring')

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        try:
            log.debug('Closing libvirt connection')

            # Anything registered with a connection must be explicitly deregistered before the connection is closed.
            for event in self.domevents:
                self.conn.domainEventDeregisterAny(event)
            
            self.conn.close()

        except libvirt.libvirtError as e:
            log.critical(f'Could not close connection: {e}')
            sys.exit(1)

        log.debug('Connection successfully closed')
